@@ -2,21 +2,24 @@
  * Copyright (c) 2000 Apple Computer, Inc. All rights reserved.
  *
  * @APPLE_LICENSE_HEADER_START@
- *
- * The contents of this file constitute Original Code as defined in and
- * are subject to the Apple Public Source License Version 1.1 (the
- * "License").  You may not use this file except in compliance with the
- * License.  Please obtain a copy of the License at
- * http://www.apple.com/publicsource and read it before using this file.
- *
- * This Original Code and all software distributed under the License are
- * distributed on an "AS IS" basis, WITHOUT WARRANTY OF ANY KIND, EITHER
+ * 
+ * Copyright (c) 1999-2003 Apple Computer, Inc.  All Rights Reserved.
+ * 
+ * This file contains Original Code and/or Modifications of Original Code
+ * as defined in and that are subject to the Apple Public Source License
+ * Version 2.0 (the 'License'). You may not use this file except in
+ * compliance with the License. Please obtain a copy of the License at
+ * http://www.opensource.apple.com/apsl/ and read it before using this
+ * file.
+ * 
+ * The Original Code and all software distributed under the License are
+ * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
  * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
  * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
- * FITNESS FOR A PARTICULAR PURPOSE OR NON-INFRINGEMENT.  Please see the
- * License for the specific language governing rights and limitations
- * under the License.
- *
+ * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
+ * Please see the License for the specific language governing rights and
+ * limitations under the License.
+ * 
  * @APPLE_LICENSE_HEADER_END@
  */
 
@@ -35,134 +38,171 @@ includes
 #include <CoreFoundation/CoreFoundation.h>
 #include <netinet/in.h>
 #include <arpa/inet.h>
-
-#ifdef	USE_SYSTEMCONFIGURATION_PUBLIC_APIS
 #include <SystemConfiguration/SystemConfiguration.h>
-#else	/* USE_SYSTEMCONFIGURATION_PUBLIC_APIS */
-#include <SystemConfiguration/v1Compatibility.h>
-#include <SystemConfiguration/SCSchemaDefinitions.h>
-#endif	/* USE_SYSTEMCONFIGURATION_PUBLIC_APIS */
+//#include <SystemConfiguration/SCPrivate.h>      // for SCLog()
+#define SCLog
 
 #include "ppp_msg.h"
 #include "../Family/if_ppplink.h"
 #include "ppp_client.h"
-#include "ppp_option.h"
 #include "ppp_manager.h"
+#include "ppp_option.h"
 
 /* -----------------------------------------------------------------------------
 definitions
 ----------------------------------------------------------------------------- */
 
+#define CREATESERVICESETUP(a)	SCDynamicStoreKeyCreateNetworkServiceEntity(0, \
+                    kSCDynamicStoreDomainSetup, kSCCompAnyRegex, a)
+#define CREATESERVICESTATE(a)	SCDynamicStoreKeyCreateNetworkServiceEntity(0, \
+                    kSCDynamicStoreDomainState, kSCCompAnyRegex, a)
+#define CREATEPREFIXSETUP()	SCDynamicStoreKeyCreate(0, CFSTR("%@/%@/%@/"), \
+                    kSCDynamicStoreDomainSetup, kSCCompNetwork, kSCCompService)
+#define CREATEPREFIXSTATE()	SCDynamicStoreKeyCreate(0, CFSTR("%@/%@/%@/"), \
+                    kSCDynamicStoreDomainState, kSCCompNetwork, kSCCompService)
+#define CREATEGLOBALSETUP(a)	SCDynamicStoreKeyCreateNetworkGlobalEntity(0, \
+                    kSCDynamicStoreDomainSetup, a)
 
 /* -----------------------------------------------------------------------------
 Forward Declarations
 ----------------------------------------------------------------------------- */
 
-static __inline__ CFTypeRef isA_CFType(CFTypeRef obj, CFTypeID type);
-static __inline__ CFTypeRef isA_CFDictionary(CFTypeRef obj);
-static __inline__ void my_CFRelease(CFTypeRef obj);
 static CFStringRef parse_component(CFStringRef key, CFStringRef prefix);
-static u_int32_t CFStringAddrToLong(CFStringRef string);
-static boolean_t cache_notifier(SCDSessionRef session, void * arg);
 
-static void reorder_services(SCDSessionRef session);
-static int process_servicestate(SCDSessionRef session, CFStringRef serviceID);
-static int process_servicesetup(SCDSessionRef session, CFStringRef serviceID);
+static void reorder_services();
+static int process_servicestate(CFStringRef serviceID);
+static int process_servicesetup(CFStringRef serviceID);
+static void cache_notifier(SCDynamicStoreRef session, CFArrayRef changedKeys, void *info);
 
 
 /* -----------------------------------------------------------------------------
 globals
 ----------------------------------------------------------------------------- */
 
-char gLoggedInUser[32];
+CFStringRef		gLoggedInUser = NULL;
+SCDynamicStoreRef	gCfgCache = NULL;
 
 /* -----------------------------------------------------------------------------
 install the cache notification and read the initilal configured interfaces
 must be called when session cache has been setup
 ----------------------------------------------------------------------------- */
-void options_init_all(SCDSessionRef session)
+int options_init_all()
 {
-    SCDStatus		status;
-    CFStringRef         serviceID, key, prefix;
-    CFArrayRef		services;
-    u_int32_t 		i;
+    CFStringRef         key = NULL, setup = NULL;
+    CFMutableArrayRef	keys = NULL, patterns = NULL;
+    CFArrayRef		services = NULL;
+    int 		i, nb, ret = 0;
+    CFRunLoopSourceRef	rls = NULL;
+    CFStringRef         notifs[] = {
+        kSCEntNetPPP,
+        kSCEntNetModem,
+        kSCEntNetInterface,
+    	kSCEntNetIPv4,
+    	kSCEntNetIPv6,
+        kSCEntNetDNS,
+        NULL,
+    };
 
-    gLoggedInUser[0] = 0;
-    SCDConsoleUserGet(gLoggedInUser, sizeof(gLoggedInUser), 0, 0);
+    /* opens now our session to the cache */
+    if ((gCfgCache = SCDynamicStoreCreate(0, CFSTR("PPPController"), cache_notifier, 0)) == NULL)
+        goto fail;
+    
+    if ((rls = SCDynamicStoreCreateRunLoopSource(0, gCfgCache, 0)) == NULL) 
+        goto fail;
+    CFRunLoopAddSource(CFRunLoopGetCurrent(), rls, kCFRunLoopDefaultMode);
+    CFRelease(rls);
+
+    gLoggedInUser = SCDynamicStoreCopyConsoleUser(gCfgCache, 0, 0);
+    
+    keys = CFArrayCreateMutable(0, 0, &kCFTypeArrayCallBacks);
+    patterns = CFArrayCreateMutable(0, 0, &kCFTypeArrayCallBacks);
+    if (keys == NULL || patterns == NULL)
+        goto fail;
 
     /* install the notifier for the cache/setup changes */
-    key = SCDKeyCreateNetworkServiceEntity(kSCCacheDomainSetup, kSCCompAnyRegex, kSCEntNetPPP);
-    status = SCDNotifierAdd(session, key, kSCDRegexKey);
+    for (i = 0; notifs[i]; i++) {   
+        if ((key = CREATESERVICESETUP(notifs[i])) == NULL)
+            goto fail;    
+        CFArrayAppendValue(patterns, key);
+        CFRelease(key);
+    }
+    
+    /* install the notifier for the cache/state changes */
+    if ((key = CREATESERVICESTATE(kSCEntNetPPP)) == NULL)
+        goto fail;
+    CFArrayAppendValue(patterns, key);
     CFRelease(key);
-    key = SCDKeyCreateNetworkServiceEntity(kSCCacheDomainSetup, kSCCompAnyRegex, kSCEntNetModem);
-    status = SCDNotifierAdd(session, key, kSCDRegexKey);
-    CFRelease(key);
-    key = SCDKeyCreateNetworkServiceEntity(kSCCacheDomainSetup, kSCCompAnyRegex, kSCEntNetInterface);
-    status = SCDNotifierAdd(session, key, kSCDRegexKey);
-    CFRelease(key);
-    key = SCDKeyCreateNetworkServiceEntity(kSCCacheDomainSetup, kSCCompAnyRegex, kSCEntNetIPv4);
-    status = SCDNotifierAdd(session, key, kSCDRegexKey);
-    CFRelease(key);
-    key = SCDKeyCreateNetworkServiceEntity(kSCCacheDomainSetup, kSCCompAnyRegex, kSCEntNetDNS);
-    status = SCDNotifierAdd(session, key, kSCDRegexKey);
-    CFRelease(key);
-    key = SCDKeyCreateNetworkGlobalEntity(kSCCacheDomainSetup, kSCEntNetIPv4);
-    status = SCDNotifierAdd(session, key, 0);
+    
+    /* install the notifier for the global changes */
+    if ((key = CREATEGLOBALSETUP(kSCEntNetIPv4)) == NULL)
+        goto fail;
+    CFArrayAppendValue(keys, key);
     CFRelease(key);
     
     /* install the notifier for user login/logout */
-    key = SCDKeyCreateConsoleUser();
-    status = SCDNotifierAdd(session, key, 0);
-    CFRelease(key);
-    
-    /* install the notifier for the cache/state changes */
-    key = SCDKeyCreateNetworkServiceEntity(kSCCacheDomainState, kSCCompAnyRegex, kSCEntNetPPP);
-    status = SCDNotifierAdd(session, key, kSCDRegexKey);
-    CFRelease(key);
+    if ((key = SCDynamicStoreKeyCreateConsoleUser(0)) == NULL)
+        goto fail;
+    CFArrayAppendValue(keys, key);
+    CFRelease(key);    
 
-    /* let's say we want to be informed via call back for the changes */
-    SCDNotifierInformViaCallback(session, cache_notifier, NULL);
+    /* add all the notification in one chunk */
+    SCDynamicStoreSetNotificationKeys(gCfgCache, keys, patterns);
 
     /* read the initial configured interfaces */
-    prefix = SCDKeyCreate(CFSTR("%@/%@/%@/"), kSCCacheDomainSetup, kSCCompNetwork, kSCCompService);
-    key = SCDKeyCreateNetworkServiceEntity(kSCCacheDomainSetup, kSCCompAnyRegex, kSCEntNetPPP);
-    status = SCDList(session, key, kSCDRegexKey, &services);
-    if (status == SCD_OK) {
-        for (i = 0; i < CFArrayGetCount(services); i++) {
-            serviceID = parse_component(CFArrayGetValueAtIndex(services, i), prefix);
-            if (serviceID) {
-                process_servicesetup(session, serviceID);            
-                CFRelease(serviceID);
-            }
+    key = CREATESERVICESETUP(kSCEntNetPPP);
+    setup = CREATEPREFIXSETUP();
+    if (key == NULL || setup == NULL)
+        goto fail;
+        
+    services = SCDynamicStoreCopyKeyList(gCfgCache, key);
+    if (services == NULL)
+        goto done;	// no PPP service setup
+
+    nb = CFArrayGetCount(services);
+    for (i = 0; i < nb; i++) {
+        CFStringRef serviceID;
+        if (serviceID = parse_component(CFArrayGetValueAtIndex(services, i), setup)) {
+            process_servicesetup(serviceID);            
+            CFRelease(serviceID);
         }
-        my_CFRelease(services);
     }
     
-    my_CFRelease(prefix);
-    my_CFRelease(key);
-    reorder_services(session);
+    reorder_services();
     ppp_postupdatesetup();
     //ppp_printlist();
+done:    
+    if (services) CFRelease(services);
+    if (key) CFRelease(key);
+    if (setup) CFRelease(setup);
+    if (keys) CFRelease(keys);
+    if (patterns) CFRelease(patterns);
+    return ret;
+fail:
+    SCLog(TRUE, LOG_ERR, CFSTR("PPPController options_init_all : allocation failed, error = %s\n"),
+                SCErrorString(SCError()));
+    if (gCfgCache) CFRelease(gCfgCache);
+    ret = 1;
+    goto done;
 }
 
 /* -----------------------------------------------------------------------------
 install the cache notification and read the initilal configured interfaces
 must be called when session cache has been setup
 ----------------------------------------------------------------------------- */
-int process_servicesetup(SCDSessionRef session, CFStringRef serviceID)
+int process_servicesetup(CFStringRef serviceID)
 {
-    CFDictionaryRef	service = NULL, subservice = NULL, interface;
+    CFDictionaryRef	service = NULL, interface;
     CFStringRef         subtype = NULL, iftype = NULL;
     struct ppp 		*ppp;
 
     ppp = ppp_findbyserviceID(serviceID);
 
-    interface = getEntity(session, kSCCacheDomainSetup, serviceID, kSCEntNetInterface);
+    interface = copyEntity(kSCDynamicStoreDomainSetup, serviceID, kSCEntNetInterface);
     if (interface)
         iftype = CFDictionaryGetValue(interface, kSCPropNetInterfaceType);
 
     if (!interface || 
-        (iftype && (CFStringCompare(iftype, kSCValNetInterfaceTypePPP, 0) != kCFCompareEqualTo))) {
+        (iftype && !CFEqual(iftype, kSCValNetInterfaceTypePPP))) {
         // check to see if service has disappear
         if (ppp) {
             //SCDLog(LOG_INFO, CFSTR("Service has disappear : %@"), serviceID);
@@ -171,7 +211,7 @@ int process_servicesetup(SCDSessionRef session, CFStringRef serviceID)
         goto done;
     }
 
-    service = getEntity(session, kSCCacheDomainSetup, serviceID, kSCEntNetPPP);
+    service = copyEntity(kSCDynamicStoreDomainSetup, serviceID, kSCEntNetPPP);
     if (!service)	// should we allow creating PPP configuration without PPP dictionnary ?
         goto done;
 
@@ -182,7 +222,7 @@ int process_servicesetup(SCDSessionRef session, CFStringRef serviceID)
     
     //SCDLog(LOG_INFO, CFSTR("change appears, subtype = %d, serviceID = %@\n"), subtype, serviceID);
 
-    if (ppp && CFStringCompare(subtype, ppp->subtypeRef, 0) != kCFCompareEqualTo) {
+    if (ppp && !CFEqual(subtype, ppp->subtypeRef)) {
         // subtype has changed
         ppp_dispose(ppp);
         ppp = 0;
@@ -198,9 +238,8 @@ int process_servicesetup(SCDSessionRef session, CFStringRef serviceID)
     ppp_updatesetup(ppp, service);
        
 done:            
-    my_CFRelease(interface);
-    my_CFRelease(service);
-    my_CFRelease(subservice);
+    if (interface) CFRelease(interface);
+    if (service) CFRelease(service);
     return 0;
 }
 
@@ -208,7 +247,7 @@ done:
 install the cache notification and read the initilal configured interfaces
 must be called when session cache has been setup
 ----------------------------------------------------------------------------- */
-int process_servicestate(SCDSessionRef session, CFStringRef serviceID)
+int process_servicestate(CFStringRef serviceID)
 {
     CFDictionaryRef	service = NULL;
     struct ppp 		*ppp;
@@ -217,7 +256,7 @@ int process_servicestate(SCDSessionRef session, CFStringRef serviceID)
     if (!ppp)
         return 0;
 
-    service = getEntity(session, kSCCacheDomainState, serviceID, kSCEntNetPPP);
+    service = copyEntity(kSCDynamicStoreDomainState, serviceID, kSCEntNetPPP);
     if (!service)
         return 0;
 
@@ -229,34 +268,28 @@ int process_servicestate(SCDSessionRef session, CFStringRef serviceID)
 
 /* -----------------------------------------------------------------------------
 ----------------------------------------------------------------------------- */
-void reorder_services(SCDSessionRef session)
+void reorder_services()
 {
     CFDictionaryRef	ip_dict = NULL;
     CFStringRef		key, serviceID;
     CFArrayRef		serviceorder;
-    SCDHandleRef	handle = NULL;
-    int 		i;
+    int 		i, nb;
     struct ppp		*ppp;
 
-    key = SCDKeyCreateNetworkGlobalEntity(kSCCacheDomainSetup, kSCEntNetIPv4);
+    key = CREATEGLOBALSETUP(kSCEntNetIPv4);
     if (key) {
-        if (SCDGet(session, key, &handle) == SCD_OK) {
-            ip_dict  = SCDHandleGetData(handle);
-            if (ip_dict) {
-                serviceorder = CFDictionaryGetValue(ip_dict, kSCPropNetServiceOrder);        
-                if (serviceorder) {
-                    for (i = 0; i < CFArrayGetCount(serviceorder); i++) {
-    
-                        serviceID = CFArrayGetValueAtIndex(serviceorder, i);
-                        
-                        ppp = ppp_findbyserviceID(serviceID);
-                        if (ppp) {
-                            ppp_setorder(ppp, 0xffff);
-                        }
-                    }
+        ip_dict = (CFDictionaryRef)SCDynamicStoreCopyValue(gCfgCache, key);
+        if (ip_dict) {
+            serviceorder = CFDictionaryGetValue(ip_dict, kSCPropNetServiceOrder);        
+            if (serviceorder) {
+  	        nb = CFArrayGetCount(serviceorder);
+	        for (i = 0; i < nb; i++) {
+                    serviceID = CFArrayGetValueAtIndex(serviceorder, i);                    
+                    if (ppp = ppp_findbyserviceID(serviceID))
+                        ppp_setorder(ppp, 0xffff);
                 }
             }
-            SCDHandleRelease(handle);
+            CFRelease(ip_dict);
         }
         CFRelease(key);
     }
@@ -265,113 +298,84 @@ void reorder_services(SCDSessionRef session)
 /* -----------------------------------------------------------------------------
 the configd cache/setup has changed
 ----------------------------------------------------------------------------- */
-boolean_t cache_notifier(SCDSessionRef session, void * arg)
+void cache_notifier(SCDynamicStoreRef session, CFArrayRef changedKeys, void *info)
 {
-    CFArrayRef          changes = NULL;
-    CFStringRef		prefixsetup = NULL, prefixstate = NULL;
-    SCDStatus           status;
-    u_long		i, doreorder = 0, dopostsetup = 0;
-    char 		str[32];
+    CFStringRef		setup, state, userkey, ipkey;
+    u_long		i, nb, doreorder = 0, dopostsetup = 0;
 
-    status = SCDNotifierGetChanges(session, &changes);
-    if (status != SCD_OK || changes == NULL) 
-        return TRUE;
+    //SCLog(TRUE, LOG_ERR, CFSTR("PPPController cache_notifier \n"));
     
-    prefixsetup = SCDKeyCreate(CFSTR("%@/%@/%@/"), kSCCacheDomainSetup, kSCCompNetwork, kSCCompService);
-    prefixstate = SCDKeyCreate(CFSTR("%@/%@/%@/"), kSCCacheDomainState, kSCCompNetwork, kSCCompService);
-
-    //SCDLog(LOG_INFO, CFSTR("ppp_setup_change Changes: %@"), changes);
+    if (changedKeys == NULL) 
+        return;
     
-    for (i = 0; i < CFArrayGetCount(changes); i++) {
+    setup = CREATEPREFIXSETUP();        
+    state = CREATEPREFIXSTATE();
+    userkey = SCDynamicStoreKeyCreateConsoleUser(0);
+    ipkey = CREATEGLOBALSETUP(kSCEntNetIPv4);
+    
+    if (setup == NULL || state == NULL || userkey == NULL || ipkey == NULL) {
+        SCLog(TRUE, LOG_ERR, CFSTR("PPPController cache_notifier : can't allocate keys\n"));
+        goto done;
+    }
 
-        CFStringRef	change = NULL, serviceID = NULL, globalip = NULL, user = NULL;
+    nb = CFArrayGetCount(changedKeys);
+    for (i = 0; i < nb; i++) {
+
+        CFStringRef	change, serviceID;
         
-        change = CFArrayGetValueAtIndex(changes, i);
+        change = CFArrayGetValueAtIndex(changedKeys, i);
 
         // --------- Check for change of console user --------- 
-        user = SCDKeyCreateConsoleUser();
-        if (user && CFStringCompare(change, user, 0) == kCFCompareEqualTo) {
-            str[0] = 0;
-            status = SCDConsoleUserGet(str, sizeof(str), 0, 0);
-            if (status == SCD_OK)
+        if (CFEqual(change, userkey)) {
+            CFStringRef olduser = gLoggedInUser;
+            gLoggedInUser = SCDynamicStoreCopyConsoleUser(session, 0, 0);
+            if (gLoggedInUser == 0)
+                ppp_logout();	// key disappeared, user logged out
+            else if (olduser == 0)
                 ppp_login();	// key appeared, user logged in
             else
-                ppp_logout();	// key disappeared, user logged out
-            
-            strncpy(gLoggedInUser, str, sizeof(gLoggedInUser));
-            CFRelease(user);
+                ppp_logswitch();	// key changed, user has switched
+            if (olduser) CFRelease(olduser);
             continue;
         }
-	my_CFRelease(user);
 
         // ---------  Check for change in service order --------- 
-        globalip = SCDKeyCreateNetworkGlobalEntity(kSCCacheDomainSetup, kSCEntNetIPv4);
-        if (globalip && CFStringCompare(change, globalip, 0) == kCFCompareEqualTo) {
+        if (CFEqual(change, ipkey)) {
             // can't just reorder the list now 
             // because the list may contain service not already created
             doreorder = 1;
-            CFRelease(globalip);
             continue;
         }
-	my_CFRelease(globalip);
 
         // --------- Check for change in other entities (state or setup) --------- 
-        serviceID = parse_component(change, prefixsetup);
+        serviceID = parse_component(change, setup);
         if (serviceID) {
-            process_servicesetup(session, serviceID);
+            process_servicesetup(serviceID);
             CFRelease(serviceID);
             dopostsetup = 1;
             continue;
         }
         
-        serviceID = parse_component(change, prefixstate);
+        serviceID = parse_component(change, state);
         if (serviceID) {
-            process_servicestate(session, serviceID);
+            process_servicestate(serviceID);
             CFRelease(serviceID);
             continue;
         }
-        
     }
 
-    my_CFRelease(prefixsetup);
-    my_CFRelease(prefixstate);
-    my_CFRelease(changes);
-
     if (doreorder) {
-        reorder_services(session);
+        reorder_services();
         //ppp_printlist();
     }
     if (dopostsetup)
         ppp_postupdatesetup();
-    return TRUE;
-}
 
-/* -----------------------------------------------------------------------------
------------------------------------------------------------------------------ */
-static __inline__ CFTypeRef isA_CFType(CFTypeRef obj, CFTypeID type)
-{
-    if (obj == NULL)
-        return (NULL);
-
-    if (CFGetTypeID(obj) != type)
-        return (NULL);
-    return (obj);
-}
-
-/* -----------------------------------------------------------------------------
------------------------------------------------------------------------------ */
-static __inline__ CFTypeRef isA_CFDictionary(CFTypeRef obj)
-{
-    return (isA_CFType(obj, CFDictionaryGetTypeID()));
-}
-
-
-/* -----------------------------------------------------------------------------
------------------------------------------------------------------------------ */
-static __inline__ void my_CFRelease(CFTypeRef obj)
-{
-    if (obj)
-        CFRelease(obj);
+done:
+    if (setup) CFRelease(setup);
+    if (state) CFRelease(state);
+    if (userkey) CFRelease(userkey);
+    if (ipkey) CFRelease(ipkey);
     return;
 }
 
@@ -406,32 +410,95 @@ static CFStringRef parse_component(CFStringRef key, CFStringRef prefix)
 
 /* -----------------------------------------------------------------------------
 ----------------------------------------------------------------------------- */
-CFTypeRef getEntity(SCDSessionRef session, CFStringRef domain, CFStringRef serviceID, CFStringRef entity)
+CFDictionaryRef copyService(CFStringRef domain, CFStringRef serviceID)
+{
+    CFTypeRef		data = NULL;
+    CFMutableDictionaryRef	service = NULL;
+    CFStringRef		key = NULL;
+    int			i;
+    CFStringRef         copy[] = {
+        kSCEntNetPPP,
+        kSCEntNetModem,
+        kSCEntNetInterface,
+    	kSCEntNetIPv4,
+    	kSCEntNetIPv6,
+        kSCEntNetDNS,
+        kSCEntNetL2TP,
+        kSCEntNetPPTP,
+        NULL,
+    };
+
+    key = SCDynamicStoreKeyCreate(0, CFSTR("%@/%@/%@/%@"), domain, kSCCompNetwork, kSCCompService, serviceID);
+    if (key == 0)
+        goto fail;
+        
+    data = SCDynamicStoreCopyValue(gCfgCache, key);
+    if (data == 0)
+        goto fail;
+        
+    CFRelease(key);
+        
+    service = CFDictionaryCreateMutableCopy(NULL, 0, data);
+    if (service == 0)
+        goto fail;
+        
+    CFRelease(data);
+
+    for (i = 0; copy[i]; i++) {   
+        data = copyEntity(domain, serviceID, copy[i]);
+        if (data) {
+        
+            CFDictionaryAddValue(service, copy[i], data);
+            CFRelease(data);
+        }
+    }
+
+    return service;
+
+fail:
+    if (key) 
+        CFRelease(key);
+    if (data)
+        CFRelease(data);
+    if (service)
+        CFRelease(service);
+    return 0;
+}
+
+/* -----------------------------------------------------------------------------
+----------------------------------------------------------------------------- */
+CFDictionaryRef copyEntity(CFStringRef domain, CFStringRef serviceID, CFStringRef entity)
 {
     CFTypeRef		data = NULL;
     CFStringRef		key;
-    SCDHandleRef	handle = NULL;
 
-    if (entity) 
-        key = SCDKeyCreateNetworkServiceEntity(domain, serviceID, entity);
-    else 
-        key = SCDKeyCreate(CFSTR("%@/%@/%@/%@"), domain, kSCCompNetwork, kSCCompService, serviceID);
+    if (serviceID)
+        key = SCDynamicStoreKeyCreateNetworkServiceEntity(0, domain, serviceID, entity);
+    else
+        key = SCDynamicStoreKeyCreateNetworkGlobalEntity(0, domain, entity);
+
     if (key) {
-        if (SCDGet(session, key, &handle) == SCD_OK) {
-            data  = SCDHandleGetData(handle);
-            if (data) {
-                if (CFGetTypeID(data) == CFDictionaryGetTypeID())	
-                    data = (CFTypeRef)CFDictionaryCreateMutableCopy(NULL, 0, data);
-                else if (CFGetTypeID(data) == CFStringGetTypeID())
-                    data = (CFTypeRef)CFStringCreateCopy(NULL, data);
-            }
-            SCDHandleRelease(handle);
-        }
+        data = SCDynamicStoreCopyValue(gCfgCache, key);
         CFRelease(key);
     }
     return data;
 }
 
+/* -----------------------------------------------------------------------------
+----------------------------------------------------------------------------- */
+int existEntity(CFStringRef domain, CFStringRef serviceID, CFStringRef entity)
+{
+    CFTypeRef		data;
+
+    data = copyEntity(domain, serviceID, entity);
+    if (data) {
+        CFRelease(data);
+        return 1;
+    }
+    
+    return 0;
+}
+
 /* -----------------------------------------------------------------------------
 get a string from the dictionnary, in service/property
 ----------------------------------------------------------------------------- */
@@ -445,7 +512,7 @@ int getString(CFDictionaryRef service, CFStringRef property, u_char *str, u_int1
     if (ref) {
         if (CFGetTypeID(ref) == CFStringGetTypeID()) {
            CFStringGetCString((CFStringRef)ref, str, maxlen, kCFStringEncodingUTF8);
-            return 0;
+            return 1;
         }
         else if (CFGetTypeID(ref) == CFDataGetTypeID()) {
            string = CFStringCreateWithCharacters(NULL, 
@@ -453,11 +520,11 @@ int getString(CFDictionaryRef service, CFStringRef property, u_char *str, u_int1
            if (string) {
                 CFStringGetCString(string, str, maxlen, kCFStringEncodingUTF8);
                 CFRelease(string);
-            	return 0;
+            	return 1;
             }
         }
     }
-    return -1;
+    return 0;
 }
 
 /* -----------------------------------------------------------------------------
@@ -468,60 +535,63 @@ int getNumber(CFDictionaryRef dict, CFStringRef property, u_int32_t *outval)
     CFNumberRef		ref;
 
     ref  = CFDictionaryGetValue(dict, property);
-    if (ref && (CFNumberGetType(ref) == kCFNumberSInt32Type)) {
+    if (ref && (CFGetTypeID(ref) == CFNumberGetTypeID())) {
         CFNumberGetValue(ref, kCFNumberSInt32Type, outval);
-        return 0;
+        return 1;
     }
-    return -1;
+    return 0;
 }
+
 /* -----------------------------------------------------------------------------
 ----------------------------------------------------------------------------- */
-int getNumberFromEntity(SCDSessionRef session, CFStringRef domain, CFStringRef serviceID, 
+int getNumberFromEntity(CFStringRef domain, CFStringRef serviceID, 
         CFStringRef entity, CFStringRef property, u_int32_t *outval)
 {
-    CFTypeRef		data = NULL;
-    CFStringRef		key;
-    SCDHandleRef	handle = NULL;
-    int 		err = -1;
+    CFTypeRef		data;
+    int 		ok = 0;
 
-    key = SCDKeyCreate(CFSTR("%@/%@/%@/%@/%@"), domain, kSCCompNetwork, kSCCompService, serviceID, entity);
-    if (key) {
-        if (SCDGet(session, key, &handle) == SCD_OK) {
-            data  = SCDHandleGetData(handle);
-            if (data) 
-                err = getNumber(data, property, outval);
-            SCDHandleRelease(handle);
-        }
-        CFRelease(key);
+    if (data = copyEntity(domain, serviceID, entity)) {
+        ok = getNumber(data, property, outval);
+        CFRelease(data);
     }
-    return err;
+    return ok;
 }
 
 /* -----------------------------------------------------------------------------
 ----------------------------------------------------------------------------- */
-int getStringFromEntity(SCDSessionRef session, CFStringRef domain, CFStringRef serviceID, 
+int getStringFromEntity(CFStringRef domain, CFStringRef serviceID, 
         CFStringRef entity, CFStringRef property, u_char *str, u_int16_t maxlen)
 {
-    CFTypeRef		data = NULL;
-    CFStringRef		key;
-    SCDHandleRef	handle = NULL;
-    int 		err = -1;
+    CFTypeRef		data;
+    int 		ok = 0;
 
-    if (serviceID)
-        key = SCDKeyCreate(CFSTR("%@/%@/%@/%@/%@"), domain, kSCCompNetwork, kSCCompService, serviceID, entity);
-    else
-        key = SCDKeyCreateNetworkGlobalEntity(domain, entity);
-    if (key) {
-        if (SCDGet(session, key, &handle) == SCD_OK) {
-            data  = SCDHandleGetData(handle);
-            if (data) {
-                err = getString(data, property, str, maxlen);
-            }
-            SCDHandleRelease(handle);
-        }
-        CFRelease(key);
+    data = copyEntity(domain, serviceID, entity);
+    if (data) {
+        ok = getString(data, property, str, maxlen);
+        CFRelease(data);
     }
-    return err;
+    return ok;
+}
+
+/* -----------------------------------------------------------------------------
+----------------------------------------------------------------------------- */
+CFStringRef copyCFStringFromEntity(CFStringRef domain, CFStringRef serviceID, 
+        CFStringRef entity, CFStringRef property)
+{
+    CFTypeRef		data;
+    CFStringRef		string, ret = 0;
+
+    data = copyEntity(domain, serviceID, entity);
+    if (data) {
+        string  = CFDictionaryGetValue(data, property);
+        if (string && (CFGetTypeID(string) == CFStringGetTypeID())) {
+            CFRetain(string);
+            ret = string;
+        }
+
+        CFRelease(data);
+    }
+    return ret; 
 }
 
 /* -----------------------------------------------------------------------------
@@ -541,54 +611,21 @@ u_int32_t CFStringAddrToLong(CFStringRef string)
 
 /* -----------------------------------------------------------------------------
 ----------------------------------------------------------------------------- */
-int getAddressFromEntity(SCDSessionRef session, CFStringRef domain, CFStringRef serviceID, 
+int getAddressFromEntity(CFStringRef domain, CFStringRef serviceID, 
         CFStringRef entity, CFStringRef property, u_int32_t *outval)
 {
-    CFTypeRef		data = NULL;
-    CFStringRef		key;
-    SCDHandleRef	handle = NULL;
-    int 		err = -1;
+    CFTypeRef		data;
+    int 		ok = 0;
     CFArrayRef		array;
 
-    key = SCDKeyCreate(CFSTR("%@/%@/%@/%@/%@"), domain, kSCCompNetwork, kSCCompService, serviceID, entity);
-    if (key) {
-        if (SCDGet(session, key, &handle) == SCD_OK) {
-            data  = SCDHandleGetData(handle);
-            if (data) {
-                array = CFDictionaryGetValue(data, property);
-                if (array && CFArrayGetCount(array)) {
-                    *outval = CFStringAddrToLong(CFArrayGetValueAtIndex(array, 0));
-                    err = 0;
-                }
-            }
-            SCDHandleRelease(handle);
+    data = copyEntity(domain, serviceID, entity);
+    if (data) {
+        array = CFDictionaryGetValue(data, property);
+        if (array && CFArrayGetCount(array)) {
+            *outval = CFStringAddrToLong(CFArrayGetValueAtIndex(array, 0));
+            ok = 1;
         }
-        CFRelease(key);
+        CFRelease(data);
     }
-    return err;
+    return ok;
 }
-
-
-/* -----------------------------------------------------------------------------
------------------------------------------------------------------------------ */
-int getServiceName(SCDSessionRef session, CFStringRef serviceID, u_char *str, u_int16_t maxlen)
-{
-    CFTypeRef		data = NULL;
-    CFStringRef		key;
-    SCDHandleRef	handle = NULL;
-    int 		err = -1;
-
-    key = SCDKeyCreate(CFSTR("%@/%@/%@/%@"), kSCCacheDomainSetup, kSCCompNetwork, kSCCompService, serviceID);
-    if (key) {
-        if (SCDGet(session, key, &handle) == SCD_OK) {
-            data  = SCDHandleGetData(handle);
-            if (data) {
-                err = getString(data, kSCPropUserDefinedName, str, maxlen);
-            }
-            SCDHandleRelease(handle);
-        }
-        CFRelease(key);
-    }
-    return err;
-}
-
