@@ -1,3 +1,27 @@
+/*
+ * Copyright (c) 2003 Apple Computer, Inc. All rights reserved.
+ *
+ * @APPLE_LICENSE_HEADER_START@
+ * 
+ * Copyright (c) 1999-2003 Apple Computer, Inc.  All Rights Reserved.
+ * 
+ * This file contains Original Code and/or Modifications of Original Code
+ * as defined in and that are subject to the Apple Public Source License
+ * Version 2.0 (the 'License'). You may not use this file except in
+ * compliance with the License. Please obtain a copy of the License at
+ * http://www.opensource.apple.com/apsl/ and read it before using this
+ * file.
+ * 
+ * The Original Code and all software distributed under the License are
+ * distributed on an 'AS IS' basis, WITHOUT WARRANTY OF ANY KIND, EITHER
+ * EXPRESS OR IMPLIED, AND APPLE HEREBY DISCLAIMS ALL SUCH WARRANTIES,
+ * INCLUDING WITHOUT LIMITATION, ANY WARRANTIES OF MERCHANTABILITY,
+ * FITNESS FOR A PARTICULAR PURPOSE, QUIET ENJOYMENT OR NON-INFRINGEMENT.
+ * Please see the License for the specific language governing rights and
+ * limitations under the License.
+ * 
+ * @APPLE_LICENSE_HEADER_END@
+ */
 /*
  * ccp.c - PPP Compression Control Protocol.
  *
@@ -25,7 +49,7 @@
  * OR MODIFICATIONS.
  */
 
-#define RCSID	"$Id: ccp.c,v 1.3 2002/05/23 01:09:19 callie Exp $"
+#define RCSID	"$Id: ccp.c,v 1.7 2003/08/14 00:00:29 callie Exp $"
 
 #include <stdlib.h>
 #include <string.h>
@@ -33,78 +57,147 @@
 #include "pppd.h"
 #include "fsm.h"
 #include "ccp.h"
-#ifdef MPPE
-#include "mppe.h"
-#endif
 //#include <net/ppp-comp.h>
 #include <ppp_comp.h>
 
+#ifdef MPPE
+#include "chap_ms.h"	/* mppe_xxxx_key */
+#include "lcp.h"	/* lcp_close(), lcp_fsm */
+#endif
+
 static const char rcsid[] = RCSID;
 
+/*
+ * Unfortunately there is a bug in zlib which means that using a
+ * size of 8 (window size = 256) for Deflate compression will cause
+ * buffer overruns and kernel crashes in the deflate module.
+ * Until this is fixed we only accept sizes in the range 9 .. 15.
+ * Thanks to James Carlson for pointing this out.
+ */
+#define DEFLATE_MIN_WORKS	9
+
 /*
  * Command-line options.
  */
 static int setbsdcomp __P((char **));
 static int setdeflate __P((char **));
+static char bsd_value[8];
+static char deflate_value[8];
+
+/*
+ * Option variables.
+ */
+#ifdef MPPE
+bool refuse_mppe_stateful = 1;		/* Allow stateful mode? */
+#endif
 
 static option_t ccp_option_list[] = {
     { "noccp", o_bool, &ccp_protent.enabled_flag,
       "Disable CCP negotiation" },
     { "-ccp", o_bool, &ccp_protent.enabled_flag,
-      "Disable CCP negotiation" },
+      "Disable CCP negotiation", OPT_ALIAS },
+
     { "bsdcomp", o_special, (void *)setbsdcomp,
-      "Request BSD-Compress packet compression" },
+      "Request BSD-Compress packet compression",
+      OPT_PRIO | OPT_A2STRVAL | OPT_STATIC, bsd_value },
     { "nobsdcomp", o_bool, &ccp_wantoptions[0].bsd_compress,
-      "don't allow BSD-Compress", OPT_A2COPY,
+      "don't allow BSD-Compress", OPT_PRIOSUB | OPT_A2CLR,
       &ccp_allowoptions[0].bsd_compress },
     { "-bsdcomp", o_bool, &ccp_wantoptions[0].bsd_compress,
-      "don't allow BSD-Compress", OPT_A2COPY,
+      "don't allow BSD-Compress", OPT_ALIAS | OPT_PRIOSUB | OPT_A2CLR,
       &ccp_allowoptions[0].bsd_compress },
-    { "deflate", 1, (void *)setdeflate,
-      "request Deflate compression" },
+
+    { "deflate", o_special, (void *)setdeflate,
+      "request Deflate compression",
+      OPT_PRIO | OPT_A2STRVAL | OPT_STATIC, deflate_value },
     { "nodeflate", o_bool, &ccp_wantoptions[0].deflate,
-      "don't allow Deflate compression", OPT_A2COPY,
+      "don't allow Deflate compression", OPT_PRIOSUB | OPT_A2CLR,
       &ccp_allowoptions[0].deflate },
     { "-deflate", o_bool, &ccp_wantoptions[0].deflate,
-      "don't allow Deflate compression", OPT_A2COPY,
+      "don't allow Deflate compression", OPT_ALIAS | OPT_PRIOSUB | OPT_A2CLR,
       &ccp_allowoptions[0].deflate },
+
     { "nodeflatedraft", o_bool, &ccp_wantoptions[0].deflate_draft,
       "don't use draft deflate #", OPT_A2COPY,
       &ccp_allowoptions[0].deflate_draft },
+
     { "predictor1", o_bool, &ccp_wantoptions[0].predictor_1,
-      "request Predictor-1", 1, &ccp_allowoptions[0].predictor_1 },
+      "request Predictor-1", OPT_PRIO | 1 },
     { "nopredictor1", o_bool, &ccp_wantoptions[0].predictor_1,
-      "don't allow Predictor-1", OPT_A2COPY,
+      "don't allow Predictor-1", OPT_PRIOSUB | OPT_A2CLR,
       &ccp_allowoptions[0].predictor_1 },
     { "-predictor1", o_bool, &ccp_wantoptions[0].predictor_1,
-      "don't allow Predictor-1", OPT_A2COPY,
+      "don't allow Predictor-1", OPT_ALIAS | OPT_PRIOSUB | OPT_A2CLR,
       &ccp_allowoptions[0].predictor_1 },
+
 #ifdef MPPE
-    { "mppe-40", o_special_noarg, setmppe_40,
-      "Allow 40 bit MPPE encryption" }, 
-    { "+mppe-40", o_special_noarg, setmppe_40,
-      "Allow 40 bit MPPE encryption" }, 
-    { "nomppe-40", o_special_noarg, setnomppe_40,
-      "Disallow 40 bit MPPE encryption" }, 
-    { "-mppe-40", o_special_noarg, setnomppe_40,
-      "Disallow 40 bit MPPE encryption" }, 
-    { "mppe-128", o_special_noarg, setmppe_128,
-      "Allow 128 bit MPPE encryption" }, 
-    { "+mppe-128", o_special_noarg, setmppe_128,
-      "Allow 128 bit MPPE encryption" }, 
-    { "nomppe-128", o_special_noarg, setnomppe_128,
-      "Disallow 128 bit MPPE encryption" }, 
-    { "-mppe-128", o_special_noarg, setnomppe_128,
-      "Disallow 128 bit MPPE encryption" }, 
-    { "mppe-stateless", o_special_noarg, setmppe_stateless,
-      "Allow stateless MPPE encryption" }, 
-    { "+mppe-stateless", o_special_noarg, setmppe_stateless,
-      "Allow stateless MPPE encryption" }, 
-    { "nomppe-stateless", o_special_noarg, setnomppe_stateless,
-      "Disallow stateless MPPE encryption" }, 
-    { "-mppe-stateless", o_special_noarg, setnomppe_stateless,
-      "Disallow stateless MPPE encryption" }, 
+    /* MPPE options are symmetrical ... we only set wantoptions here */
+    { "require-mppe", o_bool, &ccp_wantoptions[0].mppe,
+      "require MPPE encryption",
+      OPT_PRIO | MPPE_OPT_40 | MPPE_OPT_128 },
+    { "+mppe", o_bool, &ccp_wantoptions[0].mppe,
+      "require MPPE encryption",
+      OPT_ALIAS | OPT_PRIO | MPPE_OPT_40 | MPPE_OPT_128 },
+    { "nomppe", o_bool, &ccp_wantoptions[0].mppe,
+      "don't allow MPPE encryption", OPT_PRIO },
+    { "-mppe", o_bool, &ccp_wantoptions[0].mppe,
+      "don't allow MPPE encryption", OPT_ALIAS | OPT_PRIO },
+
+    /* We use ccp_allowoptions[0].mppe as a junk var ... it is reset later */
+    { "require-mppe-40", o_bool, &ccp_allowoptions[0].mppe,
+      "require MPPE 40-bit encryption", OPT_PRIO | OPT_A2OR | MPPE_OPT_40,
+      &ccp_wantoptions[0].mppe },
+    { "+mppe-40", o_bool, &ccp_allowoptions[0].mppe,
+      "require MPPE 40-bit encryption", OPT_PRIO | OPT_A2OR | MPPE_OPT_40,
+      &ccp_wantoptions[0].mppe },
+    { "nomppe-40", o_bool, &ccp_allowoptions[0].mppe,
+      "don't allow MPPE 40-bit encryption",
+      OPT_PRIOSUB | OPT_A2CLRB | MPPE_OPT_40, &ccp_wantoptions[0].mppe },
+    { "-mppe-40", o_bool, &ccp_allowoptions[0].mppe,
+      "don't allow MPPE 40-bit encryption",
+      OPT_ALIAS | OPT_PRIOSUB | OPT_A2CLRB | MPPE_OPT_40,
+      &ccp_wantoptions[0].mppe },
+
+    { "require-mppe-128", o_bool, &ccp_allowoptions[0].mppe,
+      "require MPPE 128-bit encryption", OPT_PRIO | OPT_A2OR | MPPE_OPT_128,
+      &ccp_wantoptions[0].mppe },
+    { "+mppe-128", o_bool, &ccp_allowoptions[0].mppe,
+      "require MPPE 128-bit encryption",
+      OPT_ALIAS | OPT_PRIO | OPT_A2OR | MPPE_OPT_128,
+      &ccp_wantoptions[0].mppe },
+    { "nomppe-128", o_bool, &ccp_allowoptions[0].mppe,
+      "don't allow MPPE 128-bit encryption",
+      OPT_PRIOSUB | OPT_A2CLRB | MPPE_OPT_128, &ccp_wantoptions[0].mppe },
+    { "-mppe-128", o_bool, &ccp_allowoptions[0].mppe,
+      "don't allow MPPE 128-bit encryption",
+      OPT_ALIAS | OPT_PRIOSUB | OPT_A2CLRB | MPPE_OPT_128,
+      &ccp_wantoptions[0].mppe },
+
+    /* strange one; we always request stateless, but will we allow stateful? */
+    { "mppe-stateful", o_bool, &refuse_mppe_stateful,
+      "allow MPPE stateful mode", OPT_PRIO },
+    { "nomppe-stateful", o_bool, &refuse_mppe_stateful,
+      "disallow MPPE stateful mode", OPT_PRIO | 1 },
+
+#ifdef __APPLE__
+    // for compatibility
+    { "mppe-stateless", o_bool, &refuse_mppe_stateful,
+      "disallow MPPE stateful mode", OPT_ALIAS | OPT_PRIO | 1 },
+    { "+mppe-stateless", o_bool, &refuse_mppe_stateful,
+      "disallow MPPE stateful mode", OPT_ALIAS | OPT_PRIO | 1 },
+    { "nomppe-stateless", o_bool, &refuse_mppe_stateful,
+      "allow MPPE stateful mode", OPT_ALIAS | OPT_PRIO },
+    { "-mppe-stateless", o_bool, &refuse_mppe_stateful,
+      "allow MPPE stateful mode", OPT_ALIAS | OPT_PRIO },
+    { "mppe-40", o_bool, &ccp_allowoptions[0].mppe,
+      "require MPPE 40-bit encryption", OPT_PRIO | OPT_A2OR | MPPE_OPT_40,
+      &ccp_wantoptions[0].mppe },
+    { "mppe-128", o_bool, &ccp_allowoptions[0].mppe,
+      "require MPPE 128-bit encryption",
+      OPT_ALIAS | OPT_PRIO | OPT_A2OR | MPPE_OPT_128,
+      &ccp_wantoptions[0].mppe },
 #endif
+#endif /* MPPE */
 
     { NULL }
 };
@@ -142,8 +235,11 @@ struct protent ccp_protent = {
     NULL,
     NULL,
     NULL,
+#ifdef __APPLE__
+    NULL,
     NULL,
     NULL
+#endif
 };
 
 fsm ccp_fsm[NUM_PPP];
@@ -191,7 +287,7 @@ static fsm_callbacks ccp_callbacks = {
  */
 #define ANY_COMPRESS(opt)	((opt).deflate || (opt).bsd_compress \
 				 || (opt).predictor_1 || (opt).predictor_2 \
-				 || (opt).mppe )
+				 || (opt).mppe)
 
 /*
  * Local state (mainly for handling reset-reqs and reset-acks).
@@ -240,6 +336,9 @@ setbsdcomp(argv)
 	ccp_allowoptions[0].bsd_bits = abits;
     } else
 	ccp_allowoptions[0].bsd_compress = 0;
+    slprintf(bsd_value, sizeof(bsd_value),
+	     rbits == abits? "%d": "%d,%d", rbits, abits);
+
     return 1;
 }
 
@@ -267,6 +366,14 @@ setdeflate(argv)
 		     DEFLATE_MIN_SIZE, DEFLATE_MAX_SIZE);
 	return 0;
     }
+    if (rbits == DEFLATE_MIN_SIZE || abits == DEFLATE_MIN_SIZE) {
+	if (rbits == DEFLATE_MIN_SIZE)
+	    rbits = DEFLATE_MIN_WORKS;
+	if (abits == DEFLATE_MIN_SIZE)
+	    abits = DEFLATE_MIN_WORKS;
+	warning("deflate option value of %d changed to %d to avoid zlib bug",
+	     DEFLATE_MIN_SIZE, DEFLATE_MIN_WORKS);
+    }
     if (rbits > 0) {
 	ccp_wantoptions[0].deflate = 1;
 	ccp_wantoptions[0].deflate_size = rbits;
@@ -277,10 +384,12 @@ setdeflate(argv)
 	ccp_allowoptions[0].deflate_size = abits;
     } else
 	ccp_allowoptions[0].deflate = 0;
+    slprintf(deflate_value, sizeof(deflate_value),
+	     rbits == abits? "%d": "%d,%d", rbits, abits);
+
     return 1;
 }
 
-
 /*
  * ccp_init - initialize CCP.
  */
@@ -315,16 +424,6 @@ ccp_init(unit)
     ccp_allowoptions[0].bsd_bits = BSD_MAX_BITS;
 
     ccp_allowoptions[0].predictor_1 = 1;
-#ifdef MPPE
-    ccp_wantoptions[0].mppe = 0;
-    ccp_wantoptions[0].mppe_stateless = 0;
-    ccp_wantoptions[0].mppe_40 = 0;
-    ccp_wantoptions[0].mppe_128 = 0;
-    ccp_allowoptions[0].mppe_stateless = 1;
-    ccp_allowoptions[0].mppe = 1;
-    ccp_allowoptions[0].mppe_40 = 1;
-    ccp_allowoptions[0].mppe_128 = 1;
-#endif /* MPPE*/
 }
 
 /*
@@ -399,8 +498,15 @@ ccp_input(unit, p, len)
      */
     oldstate = f->state;
     fsm_input(f, p, len);
-    if (oldstate == OPENED && p[0] == TERMREQ && f->state != OPENED)
+    if (oldstate == OPENED && p[0] == TERMREQ && f->state != OPENED) {
 	notice("Compression disabled by peer.");
+#ifdef MPPE
+	if (ccp_gotoptions[unit].mppe) {
+	    error("MPPE disabled, closing LCP");
+	    lcp_close(unit, "MPPE disabled by peer");
+	}
+#endif
+    }
 
     /*
      * If we get a terminate-ack and we're not asking for compression,
@@ -453,6 +559,14 @@ ccp_protrej(unit)
 {
     ccp_flags_set(unit, 0, 0);
     fsm_lowerdown(&ccp_fsm[unit]);
+
+#ifdef MPPE
+    if (ccp_gotoptions[unit].mppe) {
+	error("MPPE required but peer negotiation failed");
+	lcp_close(unit, "MPPE required but peer negotiation failed");
+    }
+#endif
+
 }
 
 /*
@@ -463,15 +577,102 @@ ccp_resetci(f)
     fsm *f;
 {
     ccp_options *go = &ccp_gotoptions[f->unit];
-    u_char opt_buf[256];
+    u_char opt_buf[CCP_MAX_OPTION_LENGTH];
 
     *go = ccp_wantoptions[f->unit];
     all_rejected[f->unit] = 0;
 
+#ifdef MPPE
+    if (go->mppe) {
+	ccp_options *ao = &ccp_allowoptions[f->unit];
+	int auth_mschap_bits = auth_done[f->unit];
+	int numbits;
+
+	/*
+	 * Start with a basic sanity check: mschap[v2] auth must be in
+	 * exactly one direction.  RFC 3079 says that the keys are
+	 * 'derived from the credentials of the peer that initiated the call',
+	 * however the PPP protocol doesn't have such a concept, and pppd
+	 * cannot get this info externally.  Instead we do the best we can.
+	 * NB: If MPPE is required, all other compression opts are invalid.
+	 *     So, we return right away if we can't do it.
+	 */
+
+#ifdef EAP
+        if (auth_done[f->unit] & (EAP_WITHPEER | EAP_PEER)) {
+        }
+        else
+        {
+#endif
+        	
+	/* Leave only the mschap auth bits set */
+	auth_mschap_bits &= (CHAP_MS_WITHPEER  | CHAP_MS_PEER |
+			     CHAP_MS2_WITHPEER | CHAP_MS2_PEER);
+	/* Count the mschap auths */
+	auth_mschap_bits >>= CHAP_MS_SHIFT;
+	numbits = 0;
+	do {
+	    numbits += auth_mschap_bits & 1;
+	    auth_mschap_bits >>= 1;
+	} while (auth_mschap_bits);
+	if (numbits > 1) {
+	    error("MPPE required, but auth done in both directions.");
+	    lcp_close(f->unit, "MPPE required but not available");
+	    return;
+	}
+	if (!numbits) {
+	    error("MPPE required, but MS-CHAP[v2] auth not performed.");
+	    lcp_close(f->unit, "MPPE required but not available");
+	    return;
+	}
+
+	/* LM auth not supported for MPPE */
+	if (auth_done[f->unit] & (CHAP_MS_WITHPEER | CHAP_MS_PEER)) {
+	    /* This might be noise */
+	    if (go->mppe & MPPE_OPT_40) {
+		notice("Disabling 40-bit MPPE; MS-CHAP LM not supported");
+		go->mppe &= ~MPPE_OPT_40;
+		ccp_wantoptions[f->unit].mppe &= ~MPPE_OPT_40;
+	    }
+	}
+#ifdef EAP
+        }
+#endif
+
+	/* Last check: can we actually negotiate something? */
+	if (!(go->mppe & (MPPE_OPT_40 | MPPE_OPT_128))) {
+	    /* Could be misconfig, could be 40-bit disabled above. */
+	    error("MPPE required, but both 40-bit and 128-bit disabled.");
+	    lcp_close(f->unit, "MPPE required but not available");
+	    return;
+	}
+
+	/* sync options */
+	ao->mppe = go->mppe;
+	/* MPPE is not compatible with other compression types */
+	ao->bsd_compress = go->bsd_compress = 0;
+	ao->predictor_1  = go->predictor_1  = 0;
+	ao->predictor_2  = go->predictor_2  = 0;
+	ao->deflate      = go->deflate      = 0;
+    }
+#endif /* MPPE */
+
     /*
      * Check whether the kernel knows about the various
      * compression methods we might request.
      */
+#ifdef MPPE
+    if (go->mppe) {
+	opt_buf[0] = CI_MPPE;
+	opt_buf[1] = CILEN_MPPE;
+	MPPE_OPTS_TO_CI(go->mppe, &opt_buf[2]);
+	/* Key material unimportant here. */
+	if (ccp_test(f->unit, opt_buf, CILEN_MPPE + MPPE_MAX_KEY_LEN, 0) <= 0) {
+	    error("MPPE required, but kernel has no support.");
+	    lcp_close(f->unit, "MPPE required but not available");
+	}
+    }
+#endif
     if (go->bsd_compress) {
 	opt_buf[0] = CI_BSD_COMPRESS;
 	opt_buf[1] = CILEN_BSD_COMPRESS;
@@ -479,23 +680,11 @@ ccp_resetci(f)
 	if (ccp_test(f->unit, opt_buf, CILEN_BSD_COMPRESS, 0) <= 0)
 	    go->bsd_compress = 0;
     }
-#ifdef MPPE
-    if (go->mppe) {
-        opt_buf[0] = CI_MPPE;
-        opt_buf[1] = CILEN_MPPE;
-        opt_buf[2] = 0;
-	/* keysize is 8 here */
-        BCOPY(mppe_master_send_key_40, opt_buf+3, 8);
-        BCOPY(mppe_master_recv_key_40, opt_buf+11, 8);
-        if (ccp_test(f->unit, opt_buf, (2*8)+3, 0) <= 0)
-            go->mppe = 0;
-    }
-#endif /*MPPE*/
     if (go->deflate) {
 	if (go->deflate_correct) {
 	    opt_buf[0] = CI_DEFLATE;
 	    opt_buf[1] = CILEN_DEFLATE;
-	    opt_buf[2] = DEFLATE_MAKE_OPT(DEFLATE_MIN_SIZE);
+	    opt_buf[2] = DEFLATE_MAKE_OPT(DEFLATE_MIN_WORKS);
 	    opt_buf[3] = DEFLATE_CHK_SEQUENCE;
 	    if (ccp_test(f->unit, opt_buf, CILEN_DEFLATE, 0) <= 0)
 		go->deflate_correct = 0;
@@ -503,7 +692,7 @@ ccp_resetci(f)
 	if (go->deflate_draft) {
 	    opt_buf[0] = CI_DEFLATE_DRAFT;
 	    opt_buf[1] = CILEN_DEFLATE;
-	    opt_buf[2] = DEFLATE_MAKE_OPT(DEFLATE_MIN_SIZE);
+	    opt_buf[2] = DEFLATE_MAKE_OPT(DEFLATE_MIN_WORKS);
 	    opt_buf[3] = DEFLATE_CHK_SEQUENCE;
 	    if (ccp_test(f->unit, opt_buf, CILEN_DEFLATE, 0) <= 0)
 		go->deflate_draft = 0;
@@ -556,21 +745,40 @@ ccp_addci(f, p, lenp)
 
     /*
      * Add the compression types that we can receive, in decreasing
-     * preference order.  Get the kernel to allocate the first one
-     * in case it gets Acked.
+     * preference order.
      */
+#ifdef MPPE
+    if (go->mppe) {
+	u_char opt_buf[CILEN_MPPE + MPPE_MAX_KEY_LEN];
+
+	p[0] = opt_buf[0] = CI_MPPE;
+	p[1] = opt_buf[1] = CILEN_MPPE;
+	MPPE_OPTS_TO_CI(go->mppe, &p[2]);
+	MPPE_OPTS_TO_CI(go->mppe, &opt_buf[2]);
+	BCOPY(mppe_recv_key, &opt_buf[CILEN_MPPE], MPPE_MAX_KEY_LEN);
+	res = ccp_test(f->unit, opt_buf, CILEN_MPPE + MPPE_MAX_KEY_LEN, 0);
+	if (res > 0)
+	    p += CILEN_MPPE;
+	else
+	    /* This shouldn't happen, we've already tested it! */
+	    lcp_close(f->unit, "MPPE required but not available in kernel");
+    }
+#endif
     if (go->deflate) {
 	p[0] = go->deflate_correct? CI_DEFLATE: CI_DEFLATE_DRAFT;
 	p[1] = CILEN_DEFLATE;
 	p[2] = DEFLATE_MAKE_OPT(go->deflate_size);
 	p[3] = DEFLATE_CHK_SEQUENCE;
 	for (;;) {
+	    if (go->deflate_size < DEFLATE_MIN_WORKS) {
+		go->deflate = 0;
+		break;
+	    }
 	    res = ccp_test(f->unit, p, CILEN_DEFLATE, 0);
 	    if (res > 0) {
 		p += CILEN_DEFLATE;
 		break;
-	    }
-	    if (res < 0 || go->deflate_size <= DEFLATE_MIN_SIZE) {
+	    } else if (res < 0) {
 		go->deflate = 0;
 		break;
 	    }
@@ -585,61 +793,25 @@ ccp_addci(f, p, lenp)
 	    p += CILEN_DEFLATE;
 	}
     }
-#ifdef MPPE
-    if (go->mppe) {
-        u_char opt_buf[64];
-        u_int keysize = 0;
-
-        if(!mppe_allowed)
-            go->mppe_stateless = go->mppe_40 = go->mppe_128 = 0;
-        p[0]=CI_MPPE;
-        p[1]=CILEN_MPPE;
-        p[2]=(go->mppe_stateless ? MPPE_STATELESS : 0);
-        p[3]=0;
-        p[4]=0;
-        p[5]=(go->mppe_40 ? MPPE_40BIT : 0) | (go->mppe_128 ? MPPE_128BIT : 0);
-        if(p[5] & MPPE_40BIT) {
-            keysize = 8;
-            BCOPY(mppe_master_send_key_40, opt_buf+3, keysize);
-            BCOPY(mppe_master_recv_key_40, opt_buf+11, keysize);
-        } else if(p[5] & MPPE_128BIT) {
-            keysize = 16;
-            BCOPY(mppe_master_send_key_128, opt_buf+3, keysize);
-            BCOPY(mppe_master_recv_key_128, opt_buf+19, keysize);
-        }
-        if(p[5] != 0) {
-            opt_buf[0]=CI_MPPE;
-            opt_buf[1]=CILEN_MPPE;
-            opt_buf[2] = (go->mppe_stateless) ? 1 : 0;
-            res = ccp_test(f->unit, opt_buf, (2*keysize)+3, 0);
-        } else {
-            res = -1;
-        }
-        if (res > 0) {
-                p += CILEN_MPPE;
-        }
-    }
-#endif /* MPPE*/
     if (go->bsd_compress) {
 	p[0] = CI_BSD_COMPRESS;
 	p[1] = CILEN_BSD_COMPRESS;
 	p[2] = BSD_MAKE_OPT(BSD_CURRENT_VERSION, go->bsd_bits);
-	if (p != p0) {
-	    p += CILEN_BSD_COMPRESS;	/* not the first option */
-	} else {
-	    for (;;) {
-		res = ccp_test(f->unit, p, CILEN_BSD_COMPRESS, 0);
-		if (res > 0) {
-		    p += CILEN_BSD_COMPRESS;
-		    break;
-		}
-		if (res < 0 || go->bsd_bits <= BSD_MIN_BITS) {
-		    go->bsd_compress = 0;
-		    break;
-		}
-		--go->bsd_bits;
-		p[2] = BSD_MAKE_OPT(BSD_CURRENT_VERSION, go->bsd_bits);
+	for (;;) {
+	    if (go->bsd_bits < BSD_MIN_BITS) {
+		go->bsd_compress = 0;
+		break;
 	    }
+	    res = ccp_test(f->unit, p, CILEN_BSD_COMPRESS, 0);
+	    if (res > 0) {
+		p += CILEN_BSD_COMPRESS;
+		break;
+	    } else if (res < 0) {
+		go->bsd_compress = 0;
+		break;
+	    }
+	    --go->bsd_bits;
+	    p[2] = BSD_MAKE_OPT(BSD_CURRENT_VERSION, go->bsd_bits);
 	}
     }
     /* XXX Should Predictor 2 be preferable to Predictor 1? */
@@ -680,6 +852,22 @@ ccp_ackci(f, p, len)
     ccp_options *go = &ccp_gotoptions[f->unit];
     u_char *p0 = p;
 
+#ifdef MPPE
+    if (go->mppe) {
+	u_char opt_buf[CILEN_MPPE];
+
+	opt_buf[0] = CI_MPPE;
+	opt_buf[1] = CILEN_MPPE;
+	MPPE_OPTS_TO_CI(go->mppe, &opt_buf[2]);
+	if (len < CILEN_MPPE || memcmp(opt_buf, p, CILEN_MPPE))
+	    return 0;
+	p += CILEN_MPPE;
+	len -= CILEN_MPPE;
+	/* XXX Cope with first/fast ack */
+	if (len == 0)
+	    return 1;
+    }
+#endif
     if (go->deflate) {
 	if (len < CILEN_DEFLATE
 	    || p[0] != (go->deflate_correct? CI_DEFLATE: CI_DEFLATE_DRAFT)
@@ -703,22 +891,6 @@ ccp_ackci(f, p, len)
 	    len -= CILEN_DEFLATE;
 	}
     }
-#ifdef MPPE
-    if (go->mppe) {
-        if ( len < CILEN_MPPE
-		|| p[1] != CILEN_MPPE || p[0] != CI_MPPE
-		|| p[2] != (go->mppe_stateless ? MPPE_STATELESS : 0)
-		|| p[3] != 0
-		|| p[4] != 0
-		|| p[5] != ((go->mppe_40 ? MPPE_40BIT : 0) | (go->mppe_128 ? MPPE_128BIT : 0)))
-           return 0;
-        p += CILEN_MPPE;
-        len -= CILEN_MPPE;
-        /* Cope with first/fast ack */
-        if (len == 0)
-            return 1;
-    }
-#endif /* MPPE */
     if (go->bsd_compress) {
 	if (len < CILEN_BSD_COMPRESS
 	    || p[0] != CI_BSD_COMPRESS || p[1] != CILEN_BSD_COMPRESS
@@ -773,6 +945,27 @@ ccp_nakci(f, p, len)
     memset(&no, 0, sizeof(no));
     try = *go;
 
+#ifdef MPPE
+    if (go->mppe && len >= CILEN_MPPE
+	&& p[0] == CI_MPPE && p[1] == CILEN_MPPE) {
+	no.mppe = 1;
+	/*
+	 * Peer wants us to use a different strength or other setting.
+	 * Fail if we aren't willing to use his suggestion.
+	 */
+	MPPE_CI_TO_OPTS(&p[2], try.mppe);
+	if ((try.mppe & MPPE_OPT_STATEFUL) && refuse_mppe_stateful)
+	    try.mppe = 0;
+	else if ((go->mppe & try.mppe) != try.mppe)
+	    /* Peer must have set options we didn't request (suggest) */
+	    try.mppe = 0;
+
+	if (!try.mppe) {
+	    error("MPPE required but peer negotiation failed");
+	    lcp_close(f->unit, "MPPE required but peer negotiation failed");
+	}
+    }
+#endif /* MPPE */
     if (go->deflate && len >= CILEN_DEFLATE
 	&& p[0] == (go->deflate_correct? CI_DEFLATE: CI_DEFLATE_DRAFT)
 	&& p[1] == CILEN_DEFLATE) {
@@ -782,7 +975,7 @@ ccp_nakci(f, p, len)
 	 * Stop asking for Deflate if we don't understand his suggestion.
 	 */
 	if (DEFLATE_METHOD(p[2]) != DEFLATE_METHOD_VAL
-	    || DEFLATE_SIZE(p[2]) < DEFLATE_MIN_SIZE
+	    || DEFLATE_SIZE(p[2]) < DEFLATE_MIN_WORKS
 	    || p[3] != DEFLATE_CHK_SEQUENCE)
 	    try.deflate = 0;
 	else if (DEFLATE_SIZE(p[2]) < go->deflate_size)
@@ -797,24 +990,6 @@ ccp_nakci(f, p, len)
 	}
     }
 
-#ifdef MPPE
-    if (len >= CILEN_MPPE && p[0] == CI_MPPE && p[1] == CILEN_MPPE) {
-	no.mppe = 1;
-	if((p[5] & MPPE_128BIT) == 0)
-	    try.mppe_128 = 0;
-	if((p[5] & MPPE_40BIT) == 0)
-            try.mppe_40 = 0;
-	if((p[2] & MPPE_STATELESS) == 0)
-            try.mppe_stateless = 0;
-	if(!try.mppe_128 && !try.mppe_40) {
-            no.mppe = 0;
-            try.mppe = 0;
-	}
-	p += CILEN_MPPE;
-	len -= CILEN_MPPE;
-    }
-#endif /* MPPE */
-
     if (go->bsd_compress && len >= CILEN_BSD_COMPRESS
 	&& p[0] == CI_BSD_COMPRESS && p[1] == CILEN_BSD_COMPRESS) {
 	no.bsd_compress = 1;
@@ -862,6 +1037,15 @@ ccp_rejci(f, p, len)
     if (len == 0 && all_rejected[f->unit])
 	return -1;
 
+#ifdef MPPE
+    if (go->mppe && len >= CILEN_MPPE
+	&& p[0] == CI_MPPE && p[1] == CILEN_MPPE) {
+	error("MPPE required but peer refused");
+	lcp_close(f->unit, "MPPE required but peer refused");
+	p += CILEN_MPPE;
+	len -= CILEN_MPPE;
+    }
+#endif
     if (go->deflate && len >= CILEN_DEFLATE
 	&& p[0] == (go->deflate_correct? CI_DEFLATE: CI_DEFLATE_DRAFT)
 	&& p[1] == CILEN_DEFLATE) {
@@ -887,14 +1071,6 @@ ccp_rejci(f, p, len)
 	if (!try.deflate_correct && !try.deflate_draft)
 	    try.deflate = 0;
     }
-#ifdef MPPE
-    if (go->mppe && len >= CILEN_MPPE
-	&& p[0] == CI_MPPE && p[1] == CILEN_MPPE) {
-	try.mppe = 0;
-	p += CILEN_MPPE;
-	len -= CILEN_MPPE;
-    }
-#endif /*MPPE*/
     if (go->bsd_compress && len >= CILEN_BSD_COMPRESS
 	&& p[0] == CI_BSD_COMPRESS && p[1] == CILEN_BSD_COMPRESS) {
 	if (p[2] != BSD_MAKE_OPT(BSD_CURRENT_VERSION, go->bsd_bits))
@@ -942,6 +1118,9 @@ ccp_reqci(f, p, lenp, dont_nak)
     int len, clen, type, nb;
     ccp_options *ho = &ccp_hisoptions[f->unit];
     ccp_options *ao = &ccp_allowoptions[f->unit];
+#ifdef MPPE
+    bool seen_ci_mppe = 0;
+#endif
 
     ret = CONFACK;
     retp = p0 = p;
@@ -962,6 +1141,110 @@ ccp_reqci(f, p, lenp, dont_nak)
 	    clen = p[1];
 
 	    switch (type) {
+#ifdef MPPE
+	    case CI_MPPE:
+		if (!ao->mppe || clen != CILEN_MPPE) {
+		    newret = CONFREJ;
+		    break;
+		}
+		seen_ci_mppe = 1;
+		MPPE_CI_TO_OPTS(&p[2], ho->mppe);
+		/* Nak if anything unsupported or unknown are set. */
+		if (ho->mppe & MPPE_OPT_UNSUPPORTED) {
+		    newret = CONFNAK;
+		    ho->mppe &= ~MPPE_OPT_UNSUPPORTED;
+		}
+		if (ho->mppe & MPPE_OPT_UNKNOWN) {
+		    newret = CONFNAK;
+		    ho->mppe &= ~MPPE_OPT_UNKNOWN;
+		}
+
+		/* Check state opt */
+		if (ho->mppe & MPPE_OPT_STATEFUL) {
+		   if (refuse_mppe_stateful) {
+			/*
+			 * We can Nak and request stateless, but it's a
+			 * lot easier to just assume the peer will request
+			 * it if he can do it; stateful mode is bad over
+			 * the Internet -- which is where we expect MPPE.
+			 */
+			newret = CONFREJ;
+			break;
+		    } else {
+			newret = CONFNAK;
+		    }
+		}
+
+		/* Find out which of {S,L} are set. */
+		if ((ho->mppe & MPPE_OPT_128)
+		     && (ho->mppe & MPPE_OPT_40)) {
+		    /* Both are set, negotiate the strongest. */
+		    newret = CONFNAK;
+		    if (ao->mppe & MPPE_OPT_128)
+			ho->mppe &= ~MPPE_OPT_40;
+		    else if (ao->mppe & MPPE_OPT_40)
+			ho->mppe &= ~MPPE_OPT_128;
+		    else {
+			newret = CONFREJ;
+			break;
+		    }
+		} else if (ho->mppe & MPPE_OPT_128) {
+		    if (!(ao->mppe & MPPE_OPT_128)) {
+			newret = CONFREJ;
+			break;
+		    }
+		} else if (ho->mppe & MPPE_OPT_40) {
+		    if (!(ao->mppe & MPPE_OPT_40)) {
+			newret = CONFREJ;
+			break;
+		    }
+		} else {
+		    /* Neither are set. */
+#ifdef __APPLE__
+                    /* retry with our options */
+		    newret = CONFNAK;
+                    if (ao->mppe & MPPE_OPT_128)
+			ho->mppe |= MPPE_OPT_128;
+		    if (ao->mppe & MPPE_OPT_40)
+			ho->mppe |= MPPE_OPT_40;
+#else
+		    newret = CONFREJ;
+		    break;
+#endif
+		}
+
+		/* rebuild the opts */
+		MPPE_OPTS_TO_CI(ho->mppe, &p[2]);
+		if (newret == CONFACK) {
+		    u_char opt_buf[CILEN_MPPE + MPPE_MAX_KEY_LEN];
+		    int mtu;
+
+		    BCOPY(p, opt_buf, CILEN_MPPE);
+		    BCOPY(mppe_send_key, &opt_buf[CILEN_MPPE],
+			  MPPE_MAX_KEY_LEN);
+		    if (ccp_test(f->unit, opt_buf,
+				 CILEN_MPPE + MPPE_MAX_KEY_LEN, 1) <= 0) {
+			/* This shouldn't happen, we've already tested it! */
+			error("MPPE required, but kernel has no support.");
+			lcp_close(f->unit, "MPPE required but not available");
+			newret = CONFREJ;
+			break;
+		    }
+		    /*
+		     * We need to decrease the interface MTU by MPPE_PAD
+		     * because MPPE frames **grow**.  The kernel [must]
+		     * allocate MPPE_PAD extra bytes in xmit buffers.
+		     */
+		    mtu = netif_get_mtu(f->unit);
+		    if (mtu)
+			netif_set_mtu(f->unit, mtu - MPPE_PAD);
+		    else{
+			newret = CONFREJ;
+                        }
+		}
+
+		break;
+#endif /* MPPE */
 	    case CI_DEFLATE:
 	    case CI_DEFLATE_DRAFT:
 		if (!ao->deflate || clen != CILEN_DEFLATE
@@ -975,7 +1258,7 @@ ccp_reqci(f, p, lenp, dont_nak)
 		ho->deflate_size = nb = DEFLATE_SIZE(p[2]);
 		if (DEFLATE_METHOD(p[2]) != DEFLATE_METHOD_VAL
 		    || p[3] != DEFLATE_CHK_SEQUENCE
-		    || nb > ao->deflate_size || nb < DEFLATE_MIN_SIZE) {
+		    || nb > ao->deflate_size || nb < DEFLATE_MIN_WORKS) {
 		    newret = CONFNAK;
 		    if (!dont_nak) {
 			p[2] = DEFLATE_MAKE_OPT(ao->deflate_size);
@@ -996,7 +1279,7 @@ ccp_reqci(f, p, lenp, dont_nak)
 			res = ccp_test(f->unit, p, CILEN_DEFLATE, 1);
 			if (res > 0)
 			    break;		/* it's OK now */
-			if (res < 0 || nb == DEFLATE_MIN_SIZE || dont_nak) {
+			if (res < 0 || nb == DEFLATE_MIN_WORKS || dont_nak) {
 			    newret = CONFREJ;
 			    p[2] = DEFLATE_MAKE_OPT(ho->deflate_size);
 			    break;
@@ -1008,84 +1291,6 @@ ccp_reqci(f, p, lenp, dont_nak)
 		}
 		break;
 
-#ifdef MPPE
-	    case CI_MPPE:
-		if (!ao->mppe || clen != CILEN_MPPE) {
-		    newret = CONFREJ;
-		    break;
-		}
-		if(!mppe_allowed)
-		{
-		    newret = CONFREJ;
-		    break;
-		}
-		ho->mppe = 1;
-		if((p[5]&(MPPE_40BIT|MPPE_128BIT)) == (MPPE_40BIT|MPPE_128BIT))
-		{
-		    /* if both are available, select the stronger */
-		    p[5] &= ~MPPE_40BIT;
-		    newret = CONFNAK;
-		}
-		if((p[2] & ~MPPE_STATELESS) != 0
-		   || p[3] != 0
-		   || p[4] != 0
-		   || (p[5] & ~(MPPE_40BIT | MPPE_128BIT)) != 0)
-		{
-    		    ccp_options *wo = &ccp_wantoptions[f->unit];
-		    /* not sure what they want, tell 'em what we got */
-		    p[2] &= MPPE_STATELESS;
-		    p[3] &= 0;
-		    p[4] &= 0;
-		    p[5] &= 0;
-//		    p[5] &= MPPE_40BIT | MPPE_128BIT;
-		    if(wo->mppe_40)
-		        p[5] |= MPPE_40BIT;
-		    if(wo->mppe_128)
-		        p[5] |= MPPE_128BIT;
-		    newret = CONFNAK;
-		}
-		
-		if((newret == CONFACK) || (newret == CONFNAK))
-		{
-    		    /* 
-		     * The kernel ppp driver needs the session key 
-                     * which is not sent via CCP :( 
-		     */
-		    unsigned int keysize;
-		    unsigned char opt_buf[64];
-		    opt_buf[0] = CI_MPPE;
-		    opt_buf[1] = CILEN_MPPE;
-                    opt_buf[2] = 0;
-		    if (p[2] & MPPE_STATELESS) {
-		      ho->mppe_stateless=1;
-		      opt_buf[2] = MPPE_STATELESS;
-		    }
-		    /* push in our send/receive keys */
-		    if(p[5] & MPPE_40BIT) {
-			ho->mppe_40 = 1;
-			keysize = 8;
-		        BCOPY(mppe_master_send_key_40, opt_buf+3, keysize);
-		        BCOPY(mppe_master_recv_key_40, opt_buf+11, keysize);
-		    } else if(p[5] & MPPE_128BIT) {
-			ho->mppe_128 = 1;
-			keysize = 16;
-		        BCOPY(mppe_master_send_key_128, opt_buf+3, keysize);
-		        BCOPY(mppe_master_recv_key_128, opt_buf+19, keysize);
-		    } else {
-		        ho->mppe = 0;
-		        newret = CONFREJ;
-		        break;
-		    }
-		    /* call ioctl and pass this nasty stuff to the kernel */
-		    if (ccp_test(f->unit, opt_buf, (2*keysize)+3, 1) <= 0){
-			ho->mppe = 0;
-			newret = CONFREJ;
-			break;
-		    }
-		}
-		break;
-#endif /* MPPE */
-
 	    case CI_BSD_COMPRESS:
 		if (!ao->bsd_compress || clen != CILEN_BSD_COMPRESS) {
 		    newret = CONFREJ;
@@ -1181,6 +1386,12 @@ ccp_reqci(f, p, lenp, dont_nak)
 	else
 	    *lenp = retp - p0;
     }
+#ifdef MPPE
+    if (ret == CONFREJ && ao->mppe && !seen_ci_mppe) {
+	error("MPPE required but peer negotiation failed");
+	lcp_close(f->unit, "MPPE required but peer negotiation failed");
+    }
+#endif
     return ret;
 }
 
@@ -1196,6 +1407,30 @@ method_name(opt, opt2)
     if (!ANY_COMPRESS(*opt))
 	return "(none)";
     switch (opt->method) {
+#ifdef MPPE
+    case CI_MPPE:
+    {
+	char *p = result;
+	char *q = result + sizeof(result); /* 1 past result */
+
+	slprintf(p, q - p, "MPPE ");
+	p += 5;
+	if (opt->mppe & MPPE_OPT_128) {
+	    slprintf(p, q - p, "128-bit ");
+	    p += 8;
+	}
+	if (opt->mppe & MPPE_OPT_40) {
+	    slprintf(p, q - p, "40-bit ");
+	    p += 7;
+	}
+	if (opt->mppe & MPPE_OPT_STATEFUL)
+	    slprintf(p, q - p, "stateful");
+	else
+	    slprintf(p, q - p, "stateless");
+
+	break;
+    }
+#endif
     case CI_DEFLATE:
     case CI_DEFLATE_DRAFT:
 	if (opt2 != NULL && opt2->deflate_size != opt->deflate_size)
@@ -1207,30 +1442,6 @@ method_name(opt, opt2)
 		     (opt->method == CI_DEFLATE_DRAFT? "(old#)": ""),
 		     opt->deflate_size);
 	break;
-#ifdef MPPE
-    case CI_MPPE:
-	if (opt->mppe_40) {
-	    if (opt->mppe_stateless) {
-		return "MPPE 40 bit, stateless";
-	    } else {
-		return "MPPE 40 bit, non-stateless";
-	    }
-	} else if (opt->mppe_128) {
-	    if (opt->mppe_stateless) {
-		return "MPPE 128 bit, stateless";
-	    } else {
-		return "MPPE 128 bit, non-stateless";
-	    }
-	} else {
-	    if (opt->mppe_stateless) {
-		return "MPPE unknown strength, stateless";
-	    } else {
-		return "MPPE unknown strength, stateless";
-	    }
-	}
-#endif
-    case CI_LZS:
-	return "Stac LZS";
     case CI_BSD_COMPRESS:
 	if (opt2 != NULL && opt2->bsd_bits != opt->bsd_bits)
 	    slprintf(result, sizeof(result), "BSD-Compress (%d/%d)",
@@ -1274,6 +1485,13 @@ ccp_up(f)
 	    notice("%s receive compression enabled", method_name(go, NULL));
     } else if (ANY_COMPRESS(*ho))
 	notice("%s transmit compression enabled", method_name(ho, NULL));
+#ifdef MPPE
+    if (go->mppe) {
+	BZERO(mppe_recv_key, MPPE_MAX_KEY_LEN);
+	BZERO(mppe_send_key, MPPE_MAX_KEY_LEN);
+	continue_networks(f->unit);		/* Bring up IP et al */
+    }
+#endif
 }
 
 /*
@@ -1287,6 +1505,16 @@ ccp_down(f)
 	UNTIMEOUT(ccp_rack_timeout, f);
     ccp_localstate[f->unit] = 0;
     ccp_flags_set(f->unit, 1, 0);
+#ifdef MPPE
+    if (ccp_gotoptions[f->unit].mppe) {
+	ccp_gotoptions[f->unit].mppe = 0;
+	if (lcp_fsm[f->unit].state == OPENED) {
+	    /* If LCP is not already going down, make sure it does. */
+	    error("MPPE disabled");
+	    lcp_close(f->unit, "MPPE disabled");
+	}
+    }
+#endif
 }
 
 /*
@@ -1343,6 +1571,27 @@ ccp_printpkt(p, plen, printer, arg)
 	    len -= optlen;
 	    optend = p + optlen;
 	    switch (code) {
+#ifdef MPPE
+	    case CI_MPPE:
+		if (optlen >= CILEN_MPPE) {
+		    u_char mppe_opts;
+
+		    MPPE_CI_TO_OPTS(&p[2], mppe_opts);
+		    printer(arg, "mppe %s %s %s %s %s %s%s",
+			    (p[2] & MPPE_H_BIT)? "+H": "-H",
+			    (p[5] & MPPE_M_BIT)? "+M": "-M",
+			    (p[5] & MPPE_S_BIT)? "+S": "-S",
+			    (p[5] & MPPE_L_BIT)? "+L": "-L",
+			    (p[5] & MPPE_D_BIT)? "+D": "-D",
+			    (p[5] & MPPE_C_BIT)? "+C": "-C",
+			    (mppe_opts & MPPE_OPT_UNKNOWN)? " +U": "");
+		    if (mppe_opts & MPPE_OPT_UNKNOWN)
+			printer(arg, " (%.2x %.2x %.2x %.2x)",
+				p[2], p[3], p[4], p[5]);
+		    p += CILEN_MPPE;
+		}
+		break;
+#endif
 	    case CI_DEFLATE:
 	    case CI_DEFLATE_DRAFT:
 		if (optlen >= CILEN_DEFLATE) {
@@ -1369,18 +1618,6 @@ ccp_printpkt(p, plen, printer, arg)
 		    p += CILEN_PREDICTOR_1;
 		}
 		break;
-	    case CI_MPPE:
-		if (optlen >= CILEN_MPPE) {
-		    printer(arg, "mppe %x %x %x %x",p[2],p[3],p[4],p[5]);
-		    p += CILEN_MPPE;
-		}
-		break;
-	    case CI_LZS:
-		if (optlen >= CILEN_LZS) {
-		    printer(arg, "lzs %x %x %x", p[2], p[3], p[4]);
-		    p += CILEN_LZS;
-		}
-		break;
 	    case CI_PREDICTOR_2:
 		if (optlen >= CILEN_PREDICTOR_2) {
 		    printer(arg, "predictor 2");
@@ -1439,10 +1676,16 @@ ccp_datainput(unit, pkt, len)
 	     */
 	    error("Lost compression sync: disabling compression");
 	    ccp_close(unit, "Lost compression sync");
+#ifdef MPPE
+	    /*
+	     * If we were doing MPPE, we must also take the link down.
+	     */
+	    if (ccp_gotoptions[unit].mppe) {
+		error("Too many MPPE errors, closing LCP");
+		lcp_close(unit, "Too many MPPE errors");
+	    }
+#endif
 	} else {
-	    /* MPPE/MPPC does not requires CCP_RESETREQ */
-	    if (ccp_gotoptions[f->unit].method == CI_MPPE)
-		return;
 	    /*
 	     * Send a reset-request to reset the peer's compressor.
 	     * We don't do that if we are still waiting for an
